/*
   program       : various
   
   $Source: /home/yee/C/RCS/fft.c,v $
   $Author: yee $
   $Date: 1999/02/10 16:39:47 $
   $Revision: 2.32 $
   
   This file contains Fourier transform code using the Fast Fourier Transform
   (FFT) technique to calculate the transforms.

   Note : output is in units of Hertz and not natural frequency (radians).

   nfft() performs n-dimensional fft's.
   fft() performs 1-dimensional fft's.

   Defaults to single precision if DOUBLE or FFT_DOUBLE not defined.

   Parameters :
      1) #define DOUBLE|FFT_DOUBLE if double precision
      2) #define NFFT to compile nfft()
      3) #define FFT to compile fft()
      4) #define NO_NORMALIZE or #define EXTERNAL_NORMALIZE to prevent
         normalization in inverse fft.

   Bugs :
      1) Several extra floats are allocated in the a, b arrays in
         fft_prep().

   To Do :
      1) Write rfft() which performs fft's on real arrays.
      2) nfft() and nfft_rec() has not been tested for any dimension greater
         than 2.
      3) The functions output the results in Hz.  Need to support the natural
         frequency (i.e. radians).

   All calls to sin() are precalculated within fft_prep() prior to calling
   fft_lemma() which implements the Danielson-Lanczos Lemma since
      1) fft_lemma() is recursive.
      2) the results sin() are stored in a static array to reduce the cost of
         future calls to transform functions.

   This code was written as an experiment on the usefulness of strict block
   structure from three points of view:
      1) amount of wasted memory usage (i.e. dummy variables that are
         allocated but not used for all but a small portion of the function)
      2) ease in the initial writing and debugging of the function
      3) ease in maintaining the function.

   $Log: fft.c,v $
   Revision 2.32  1999/02/10 16:39:47  yee
   *** empty log message ***

   Revision 2.31  1999/02/05 04:05:48  yee
   *** empty log message ***

   Revision 2.30  1999/01/09 16:23:22  yee
   *** empty log message ***

   Revision 2.29  1999/01/08 04:02:01  yee
   changed from BCOPY() to memcpy().

   Revision 2.28  1998/08/22 15:37:39  yee
   all geometries work!

   Copyright 1996,1997 by Conway Yee.  All rights reserved.
 */

char fft_c[] = "$Id: fft.c,v 2.32 1999/02/10 16:39:47 yee Exp $";

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "fft.h"
#include "byte.h"
#include "array.h"

#define FFT_CALC 2 /* 2^FFT_CALC is the smallest FFT calculated by
                      fft_lemma(). Lengths of 2^(FFT_CALC-1) are hardcoded.
                      Anything else, the fft_lemma() will return spurious
                      results.  fft_lemma() supports values of 1 or 2.  2
                      is more efficient. */

#define MASK1 (unsigned int)1    /* used for bitwise operations */

PRIVATE void fft_lemma(int,int,FFT_COMPLEX *,double *,double *,int);
PRIVATE int fft_prep(int,int *,MODE);

#ifdef NFFT
  PRIVATE void nfft_rec(int *,int,int,int *,FFT_COMPLEX *,int *,int *,
                        BOOLEAN *,int *,int *);
#endif NFFT
 
                      /* NOTE: maxr,a,b are maintained by fft_prep() */
PRIVATE int maxr=0;   /* number of precalc a,b (max # prev recurrence levels */
PRIVATE double *a=NULL; /* prev calc a of sin & cos reccurence formula */
PRIVATE double *b=NULL; /* prev calc b of sin & cos reccurence formula */

#define FFT_FLIP(ndim,nbit,data,step) { /* does byte swap as prep for fft */ \
          int i;    /* counter */                                            \
                                                                             \
          for(i=ndim-1;i>=0;i--) {                                           \
            int flip;      /* dummy var with index of flipped element */     \
            FFT_COMPLEX cdum; /* dummy var with flipped element */           \
            int size=sizeof(FFT_COMPLEX); /* size of a complex number */     \
                                                                             \
            FLIP(nbit,i,flip);                                               \
            if (i>flip) {                                                    \
               memcpy(&cdum,data+i*step,size);                               \
               memcpy(data+i*step,data+flip*step,size);                      \
               memcpy(data+flip*step,&cdum,size);                            \
             }                                                               \
          }                                                                  \
        }

#define LOCATE(off,ndim,eloc,ele) { /* offset of ele from 1st element */ \
          int i;   /* counter */                                         \
                                                                         \
          for(i=0,off=0;i<ndim;i++)                                      \
            off += *(ele+i) * *(eloc+i);                                 \
        }

#ifdef EXTERNAL_NORMALIZE
#define NO_NORMALIZE
#endif EXTERNAL_NORMALIZE

/*****************************************************************************/
/*****************************************************************************/

/* Performs one dimensional Fast Fourier transform (both forward and inverse)
   on a complex array. */

#ifdef FFT
PUBLIC FFT_COMPLEX *fft
(
 int ndim,              /* number of elements in transform data */
 FFT_COMPLEX *data,     /* transform data array */
 MODE mode              /* function option control */
)
{ /* fft() */
  int rlev;             /* levels to recurse (starting with FFT_CALC) */
  int nbit;             /* number of bits in ndim */

  if (ndim == 1)             /* handles trivial case of a 1-D FFT */
    return(data);                   /* The FFT a 1-D array is itself! */

  if(ndim<pow(2,FFT_CALC)) {
    fprintf(stderr,"fft : fft smaller than 2^FFT_CALC\n");
    return((FFT_COMPLEX *)NULL);
  }

  rlev = fft_prep(ndim,&nbit,mode);
  if( rlev >= 0 ) {
    FFT_FLIP(ndim,nbit,data,1);
    fft_lemma(rlev,ndim,data,a+rlev,b+rlev,1); /* Danielson-Lanczos Lemma */
  } else
    return((FFT_COMPLEX *)rlev);

#ifndef NO_NORMALIZE
  if (!(mode&FFT_FORWARD)) {   /* inverse transform. multiply by 1/ndim */
    FFT_COMPLEX *ptr = data;          /* pointer to data */
    int i;                            /* counter */

    for(i=0;i<ndim;i++,ptr++) {
      ptr->r /= ndim;
      ptr->i /= ndim;
    }
  }
#endif NO_NORMALIZE

  return(data);
} /* fft() */
#endif FFT

/*****************************************************************************/

/* Performs N-dimensional Fast Fourier transform (both forward and inverse)
   on a complex array.

   This function assumes that the data array is allocated as a contiguous
   block.

   WARNING : nfft() assumes that realloc will not barf on a NULL ptr.

   WARNING : nfft() has only been tested up to and including ndim = 2! */

#ifdef NFFT
PUBLIC FFT_COMPLEX *nfft
(
 int ndim,          /* dimensionality of transform data */
 int *dim,          /* number of elements in each dimension */
 FFT_COMPLEX *data, /* transform data array (address of 1st element) */
 MODE mode          /* function option control */
)

{ /* nfft() */
  static int *rlev;    /* # of levels to recurse in each dimension */
  static int *nbit;    /* number of bits in each dimension */
  static int maxdim;   /* prev max # of elements in each dim */
  static int *eloc;    /* coeff array used to locate specific elements */
  static int *ele;     /* dummy array used to identify specific elements */
  static BOOLEAN *fix; /* do not run through combinations in this dimension */
  static int *step;    /* interval between data elements in given direction */

  if (maxdim < ndim) {  /* set up utility arrays */
    maxdim = ndim;
    rlev = (int *)realloc(rlev,ndim*sizeof(int));
    nbit = (int *)realloc(nbit,ndim*sizeof(int));
    eloc = (int *)realloc(eloc,ndim*sizeof(int));
    ele = (int *)realloc(ele,ndim*sizeof(int));
    step  = (int *)realloc(step,ndim*sizeof(int));
    fix  = (BOOLEAN *)realloc(fix,ndim*sizeof(BOOLEAN));

    memset(fix,NUL,sizeof(BOOLEAN)*ndim);  /* zero out fix[] */
  } /* set up utility arrays */

  { /* initialize eloc[] */
    int i; /* counter */
      
    for(i=ndim-2,*(eloc+ndim-1)=1;i>=0;i--)
      *(eloc+i) = *(eloc+i+1) * *(dim+i);
  } /* initialize eloc[] */

  { /* rlev and nbit determination in each dimension */
    int i;             /* counter */

    for(i=0;i<ndim;i++) {
      rlev[i] = fft_prep(dim[i],&(nbit[i]),mode);
      if(rlev[i]<0)
        return((FFT_COMPLEX *)(rlev[i]));
    }
  } /* rlev and nbit determination in each dimension */

  { /* determine distance between elements in each dimension */
    int idim; /* dummy var for dimensionality */
    
    for(idim=ndim-1;idim>=0;idim--) {
      ele[idim] = 1;
      LOCATE(step[idim],ndim,eloc,ele);
      ele[idim] = 0;
    }
  } /* determine distance between elements in each direction */

  nfft_rec(rlev,ndim,ndim,dim,data,nbit,step,fix,eloc,ele);

#ifndef NO_NORMALIZE
  { /* if inverse tranform, normalize */
    int f; /* factor to multiply tranform by if inverse transform */
    int idim; /* dummy var for dimensionality */

    if( !(mode&FFT_FORWARD) ) {
      for(f=1,idim=0;idim<ndim;idim++)
        f *= dim[idim];

      for(idim=0;idim<f;idim++,data++) {
        data->r /= f;
        data->i /= f;
      }
    }
  } /* if inverse transform, normalize */
#endif NO_NORMALIZE

  return(data); /* success */
} /* nfft() */
#endif NFFT

/*****************************************************************************/
/************ PRIVATE support functions for the fft functions ****************/
/*****************************************************************************/

/* this function recursively performs the actual n-dimensional fft.

   WARNING: nfft_rec() has only been tested up to and including ndim=2! */

#ifdef NFFT
PRIVATE void nfft_rec
(
 int *rlev,          /* # of levels to recurse in each dimension */
 int rndim,          /* number of levels to recurse in N-dimensional space */
 int ndim,           /* dimensionality of transform data */
 int *dim,           /* number of elements in each dimension */
 FFT_COMPLEX *data,  /* transform data array (address of 1st element) */
 int *nbit,          /* number of bits in each dimension */
 int *step,          /* interval between data elements in given direction */
 BOOLEAN *fix,       /* do not run through combinations in this dimension */
 int *eloc,          /* coeff array used to locate specific elements */
 int *ele            /* dummy array used to identify specific elements */
)

{ /* nfft_rec() */
  if (rndim == 1) {
    int dir;  /* direction to do 1-D FFT */

    for(dir=ndim-1;dir>=0;dir--)
      if( !(fix[dir])) break;      /* unfixed direction found */
    ele[dir] = 0;
    
    { /* perform 1D fft */
      int off; /* offset from 1st data element to 1st element of 1d fft */

      LOCATE(off,ndim,eloc,ele);
      FFT_FLIP(dim[dir],nbit[dir],data+off,step[dir]);
      fft_lemma(rlev[dir],dim[dir],data+off,a+rlev[dir],b+rlev[dir],step[dir]);
    } /* perform 1D fft */

  } 
  else {
    int idim;            /* dummy ndim */

    rndim--;
    for(idim=0;idim<ndim;idim++) {
      if( !(fix[idim]) ) {
        fix[idim] = TRUE;
    
        { /* cycle through newly fixed dimension */
          int i; /* counter */

          for(i=0;i<dim[idim];i++) {
            ele[idim] = i;
            nfft_rec(rlev,rndim,ndim,dim,data,nbit,step,fix,eloc,ele);
          }
        } /* cycle through newly fixed dimension */

        fix[idim] = FALSE;
      }
    }
  }
} /* nfft_rec() */
#endif NFFT

/*****************************************************************************/

/* This is a recursive implementation of the Danielson-Lanczos Lemma.  This
   function is written to support elements that can be separated by an
   arbitrary distance (not necessarily consecutive memory locations) so that
   fft_lemma() can be used for n-dimensional fft's This code is based on the
   recursive algorithm in Press et al., Numerical Recipes (ed. II) with
   ideas from Pavlidis, Algorithms for Graphics and Image Processing.

   Note, fft_lemma() uses the results of fft_prep() for the parameters a and b.

   Note, fft_lemma() uses a and b which are offset from the PRIVATE arrays a
   and b.

   Note, all array lengths less than FFT_CALC are considered trivial and are
   either ignored [FFT length 2^(FFT_CALC-2)] or hard-coded
   [FFT length 2^(FFT_CALC-1)].  As written currently, only [FFT_CALC 1,2]
   is supported.  See WARNING below.

   BUGS: see BUGS section of file header.

   WARNING! This function assumes that ndim is a power of 2.

   WARNING! For efficiency, this function barfs on the trivial case of FFT
            lengths less than 2^(FFT_CALC-1) */

PRIVATE void fft_lemma
(
 int rlev,           /* number of levels still to recurse */
 int ndim,           /* length of data array */
 FFT_COMPLEX *data,  /* data array */
 double *a,          /* used in sin & cos recurrence formulae */
 double *b,          /* used in sin & cos recurrence formulae */
 int step            /* interval between elements. useful in n-dim fft. */
)

{ /* fft_lemma() */
  if(!rlev) {  /* end recursing */
#if FFT_CALC == 2
    { /* FFT with 2 elements */
      FFT_COMPLEX cdum; /* dummy variable */

      memcpy(&cdum,data+step,sizeof(FFT_COMPLEX)); /* copy data[1] to cdum */

      cminus(data+step,data,data+step);   /* data[1] done first to avoid */
      cplus(data,data,&cdum);             /* stomping on data[0] */
    } /* FFT with 2 elements */
#elif FFT_CALC == 1
    return;                               /* single element fft is itself */
#else
    return; /* default to calculate all ffts */
#endif
  } else {
    FFT_COMPLEX *even;                    /* ptr to even elements */
    FFT_COMPLEX *odd;                     /* ptr to odd elements */

    ndim >>= 1; /* half ndim */

    { /* calculate transforms of odd & even elements */
      double *ta=a-1;    /* ptr to a for next cycle of recursion */
      double *tb=b-1;    /* ptr to b for next cycle of recursion */

      rlev--;
      even = data;
      odd  = data+ndim*step;
                                                   /* ndim already halved */
      fft_lemma(rlev,ndim,even,ta,tb,step);        /* even elements */
      fft_lemma(rlev,ndim,odd,ta,tb,step);         /* odd elements */
    } /* calculate transforms of odd & even elements */

/* NOTE: the tranform of each half of the transform (odd and even) has a
         period half of the original transform
   NOTE: sin(theta+pi) = -1*sin(theta) and cos(theta+pi) = -1*cos(theta).  A
         full FFT cycles through an angle of 2*PI
   Thus, 1st & 2nd halves of the tranform can be calculated simultaneously by
   reusing multiplying sin() and cos() by -1 and reusing the values from each
   half of the tranform. */

    { /* calculate transform */ 
      int i;                     /* counter */
      lcomplex angle={1.0,0.0};  /* e^(i*theta)=cos(theta)+i*sin(theta) */
                                 /* theta=0,cos(0)=1.0,sin(0)=0.0 */

      { /* step through data */
        for(i=ndim;i;i--,odd+=step,even+=step) {  /* ndim already halved */
          lcomplex dum;  /* intermediate var */

          ctimes(&dum,odd,&angle);     /* odd element * angle */

                          /*  2nd half done first to avoid stomping on even */
          cminus(odd,even,&dum);       /* 2nd half of the data */
          cplus(even,even,&dum);       /* 1st half of the data */

          UpdateSinCosRecurrence(*a,*b,angle.r,angle.i);  /* increment angle */
        }

      } /* step through data */
    } /* calculate transform */
  } /* ndim > 1 */
} /* fft_lemma() */

/*****************************************************************************/

/* This function does the preparation needed for the Danielson-Lanczos lemma
   including
   1) precalculating the number of levels of recursion required
   2) precalculating the coefficients of the recurrence formulae used to
      calculate sin and cos

   This function performs all the maintenance on the PRIVATE variables maxr,
   a, and b.

   fft_prep() only calculates those values of a, b >= FIRST_CALC since smaller
   fft arrays are hard coded into fft_lemma() as trivial cases.

   If successful, this function returns the number of levels of recursion.

   Otherwise, it returns an error code. */

PRIVATE int fft_prep
(
 int ndim,             /* number of elements in 1D transform data */
 int *nbit,            /* number of bits in ndim */
 MODE mode             /* function option control */
)

{ /* fft_prep() */
  int rlev=0;          /* number of levels to recurse */

                            /* slides bits of ndim to the right & counts bits 
                               until all bits are slid out.  Also, it tests to
                               make sure all of the bits except the last one
                               is 0 (i.e., ndim is a power of 2) */
    
  { /* ndim bit count, rlev calculation */
    int dim;                 /* dummy variable for ndim */
    int small;               /* smallest fft not hardcoded into fft_lemma() */

    small=NINT(pow(2,FFT_CALC));
    for(*nbit=0,dim=ndim;dim&&((dim==1)||!(MASK1&dim));dim>>=1,(*nbit)++)
      if (dim==small)
        rlev = (*nbit)+1;
    if(dim) {
      fprintf(stderr,"n not a power of 2\n");
      return((int)FFT_NOT_POWER2);
    }
    (*nbit)--;
  } /* ndim bit count,rlev calculation */ 

  /* fft_lemma() treats fft arrays of length less than FFT_CALC as trivial */
  {
    int size=sizeof(double)*(rlev+1); /* size of cos & sin arrays */
                                      /* WARNING: extra bytes allocated */

    if (maxr < rlev) { /* need more levels of recursion */
      a=(double *)realloc(a,size);
      b=(double *)realloc(b,size);
    }
    /* otherwise, all a, b already precalculated. Only need to check
       the sign of the a, b arrays. */

    { /* update a, b arrays used in sin & cos recurrence formulae */
      int i;                            /* counter */
      double theta;                     /* angle */
      double *ta=a+rlev;                /* temp pointer to a arrays */ 
      double *tb=b+rlev;                /* temp pointer to b arrays */ 
      char sign= mode&FFT_FORWARD?1:-1; /* -'ve sign if inverse transform 
                                           +'ve sign if forward transform */

                                                     /* cycle through a,b */
      for(i=rlev,theta=(int)sign*TWO_PI/ndim;i;i--,theta*=2.0)
        if(i<=maxr) { /* precalculated a,b */
          ta--;                         /* a stays the same regardless */
          *(tb) = sign*fabs((*tb--));   /* sin(-1*theta) = -1*sin(theta) */

        } else {
          double dum;    /* dummy var for unused initialized cos_t, sin_t */

          InitializeSinCosRecurrence(*ta,*tb,dum,dum,1.0,0.0,theta);
          ta--;tb--;
        }
      if(rlev>maxr)
        maxr = rlev;               /* update maxr */
    } /* update a, b arrays */
  } /* precalc a, b of of recurrence formulae */

  return(rlev); /* number of levels of recursion */
} /* fft_prep() */
